from typing import Any, List
import logging

from ib_insync.ib import IB
from ib_insync.objects import BarDataList
from ib_insync.contract import Stock, ContractDetails

from .encodings import Api, Exchange, ContractType, Symbol
from .config_reader import ConfigReader


logger = logging.getLogger(__name__)


class TwsConnector():
    """Provides methods to download data from IB TWS"""

    def __init__(self) -> None:
        self.__config_reader = ConfigReader()
        self.__ib = IB()  # Create connection objec
        self.__ib.RaiseRequestErrors = True  # Enable exceptions
        logging.getLogger("ib_insync").setLevel(
            logging.WARN)  # todo: change with verbosity level option

    def connect(self) -> None:
        """Connect to TWS"""

        IP = self.__config_reader.get_config_value_single(
            section="tws_connector",
            option="ip")
        PORT = int(self.__config_reader.get_config_value_single(
            section="tws_connector",
            option="port"))
        self.__ib.connect(host=IP, port=PORT, clientId=1, readonly=True)
        logger.debug(f"Connected to TWS on {IP}:{PORT}.")

    def disconnect(self) -> None:
        """Disconnect from TWS"""

        self.__ib.disconnect()
        logger.debug("Disconnected from TWS.")

    def download_historical_quotes(
            self,
            symbol: str,
            exchange: str,
            currency: str,
            duration: str) -> BarDataList:
        """Download historical quotes for a contract from IB TWS

        :param symbol: Contracts symbol
        :type symbol: str
        :param exchange: Contracts exchange
        :type exchange: str
        :param currency: Contracts currency
        :type currency: str
        :param duration: Duration to download
        :type duration: str
        :return: 
        :rtype: ib_insync.objects.BarDataList
        """

        exchange = Exchange.encode(name=exchange, to_api=Api.IB)
        symbol = Symbol.encode(name=symbol, to_api=Api.IB)
        ib_contract = Stock(
            symbol=symbol,
            exchange=exchange,
            currency=currency)
        bar_data = self.__ib.reqHistoricalData(
            contract=ib_contract,
            endDateTime='',
            durationStr=duration,
            barSizeSetting='1 day',
            whatToShow='ADJUSTED_LAST',
            useRTH=True)
        logger.debug(
            f"Received {len(bar_data)} quotes for {exchange}_{symbol}_"
            f"{currency}_{duration.replace(' ', '')} from TWS.")
        return bar_data

    def download_contract_details(self, broker_symbol: str, exchange: str,
                                  currency: str) -> List[ContractDetails]:
        """Download details for a contract from IB TWS

        :param symbol: Contracts symbol
        :type symbol: str
        :param exchange: Contracts exchange
        :type exchange: str
        :param currency: Contracts currency
        :type currency: str
        :return: Details object for the contract from IB TWS
        :rtype: ib_insync.contract.ContractDetails
        """

        exchange = Exchange.encode(name=exchange, to_api=Api.IB)
        symbol = Symbol.encode(name=broker_symbol, to_api=Api.IB)
        ib_contract = Stock(
            symbol=symbol,
            exchange=exchange,
            currency=currency)
        details = self.__ib.reqContractDetails(ib_contract)
        self.__validate_details(
            details=details,
            broker_symbol=broker_symbol,
            exchange=exchange,
            currency=currency)
        logger.debug(
            f"Received contract details for {broker_symbol}_{exchange}_"
            f"{currency} from TWS")
        return self.__decode_details(details[0])

    def __validate_details(
            self,
            details: List[ContractDetails],
            broker_symbol: str,
            exchange: str,
            currency: str) -> None:
        if details is None:
            raise IbDetailsInvalidError(
                f"Result for {broker_symbol}_{exchange}_{currency} is None")
        elif len(details) == 0:
            raise IbDetailsInvalidError(
                f"Result for {broker_symbol}_{exchange}_{currency} is []")
        elif len(details) > 1:
            raise IbDetailsInvalidError(
                f"Multiple results for {broker_symbol}_{exchange}_{currency}")

    def __decode_details(self, details: ContractDetails) -> ContractDetails:
        details.contract.exchange = Exchange.decode(
            name=details.contract.exchange,
            from_api=Api.IB)
        details.contract.primaryExchange = Exchange.decode(
            name=details.contract.primaryExchange,
            from_api=Api.IB)
        details.stockType = ContractType.decode(
            name=details.stockType,
            from_api=Api.IB)
        return details


class IbDetailsInvalidError(Exception):
    """"Doc"""

    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)



import logging
from pathlib import Path

from .cli import cli


def my_filenamer(filename):
    new_name = filename.replace(".log.", "_") + ".log"
    return new_name


def main():
    """Docstring"""

    # Setup logging
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    console_formatter = logging.Formatter("%(message)s")
    console_handler.setFormatter(console_formatter)
    root_logger.addHandler(console_handler)

    file_handler = logging.handlers.TimedRotatingFileHandler(
        filename=Path.home() / ".barbucket/logfile.log",
        when='midnight')
    file_handler.namer = my_filenamer
    file_handler.setLevel(logging.DEBUG)
    file_formatter = logging.Formatter(
        "%(asctime)s | %(name)s | %(levelname)s | %(message)s")
    file_handler.setFormatter(file_formatter)
    root_logger.addHandler(file_handler)

    logger = logging.getLogger(__name__)
    logger.debug("---------------------------------------")
    logger.debug("Application started")

    # Run Cli
    cli()



import sqlite3
from pathlib import Path
import logging

from .config_reader import ConfigReader

logger = logging.getLogger(__name__)


class DbConnector():
    """Handles all non-specific database operations."""
    __db_path: Path
    __is_initialized: bool = False

    def __init__(self) -> None:
        self.__config_reader = ConfigReader()
        if not DbConnector.__is_initialized:
            self.__initialize_database()
            DbConnector.__is_initialized = True

    def connect(self) -> sqlite3.Connection:
        """Provides a 'connection' object for the database.

        :raises DBNotInitializedError: The database needs to be initialized 
        by an internal method before use, this is handled automatically.
        :return: Connection object to the database
        :rtype: sqlite3.Connection
        """

        if not DbConnector.__db_path.is_file():
            raise DBNotInitializedError(
                f"Database file {DbConnector.__db_path} does not exist.")
            # otherwise sqlite3.connect() would create an empty file and
            # we dont want that
        conn = sqlite3.connect(DbConnector.__db_path)
        conn.execute("""
            PRAGMA foreign_keys = 1;
        """)
        # On SQLite, PRAGMA commands are only valid for the existing connection,
        # so this needs to be part of every db connection.
        return conn

    def disconnect(self, conn: sqlite3.Connection) -> None:
        """Disconnects the connection to the database.

        :param conn: Connection object, provided from this class
        :type conn: sqlite3.Connection
        """
        conn.close()

    def __initialize_database(self) -> None:
        """Initialize database if it doesnt exist. Else skip."""

        self.__get_db_path()
        if not DbConnector.__db_path.is_file():
            self.__create_db_file()
            self.__create_db_schema()
            logger.info("Database created. Finished db initialization.")
        else:
            logger.debug("Database already exists. Finished initialization.")

    def __get_db_path(self) -> None:
        db_path = self.__config_reader.get_config_value_single(
            section="database",
            option="db_location")
        DbConnector.__db_path = Path.home() / Path(db_path)

    def __create_db_file(self) -> None:
        conn = sqlite3.connect(DbConnector.__db_path)
        conn.close()
        logger.debug(f"Created new database file {DbConnector.__db_path}")

    def __create_db_schema(self) -> None:
        conn = self.connect()
        cur = conn.cursor()

        cur.execute(
            """CREATE TABLE contracts (
                    contract_id INTEGER NOT NULL PRIMARY KEY,
                    contract_type_from_listing TEXT,
                    exchange_symbol TEXT, 
                    broker_symbol TEXT, 
                    name TEXT,
                    currency TEXT, 
                    exchange TEXT);""")

        cur.execute(
            """CREATE TABLE contract_details_ib (
                    contract_id INTEGER UNIQUE,
                    contract_type_from_details TEXT,
                    primary_exchange TEXT,
                    industry TEXT,
                    category TEXT,
                    subcategory TEXT,
                    FOREIGN KEY (contract_id)
                        REFERENCES contracts (contract_id)
                            ON UPDATE CASCADE
                            ON DELETE CASCADE,
                    UNIQUE (contract_id));""")

        cur.execute(
            """CREATE TABLE contract_details_tv (
                    contract_id INTEGER UNIQUE,
                    market_cap INTEGER,
                    avg_vol_30_in_curr INTEGER,
                    country TEXT,
                    employees INTEGER,
                    profit INTEGER,
                    revenue INTEGER,
                    FOREIGN KEY (contract_id)
                        REFERENCES contracts (contract_id)
                            ON UPDATE CASCADE
                            ON DELETE CASCADE,
                    UNIQUE (contract_id));""")

        cur.execute(
            """CREATE VIEW all_contract_info AS
                    SELECT * FROM contracts
                        LEFT JOIN contract_details_ib ON
                            contracts.contract_id = contract_details_ib.contract_id
                        LEFT JOIN contract_details_tv ON
                            contracts.contract_id = contract_details_tv.contract_id
                        LEFT JOIN quotes_status ON
                            contracts.contract_id = quotes_status.contract_id;""")

        cur.execute(
            """CREATE TABLE quotes (
                    contract_id INTEGER,
                    date TEXT,
                    open REAL,
                    high REAL, 
                    low REAL, 
                    close REAL,
                    volume REAL,
                    FOREIGN KEY (contract_id)
                        REFERENCES contracts (contract_id)
                            ON UPDATE CASCADE
                            ON DELETE CASCADE,
                    UNIQUE (contract_id, date));""")

        cur.execute(
            """CREATE TABLE quotes_status (
                    contract_id INTEGER UNIQUE,
                    status_code INTEGER,
                    status_text TEXT,
                    daily_quotes_requested_from TEXT,
                    daily_quotes_requested_till TEXT,
                    FOREIGN KEY (contract_id)
                        REFERENCES contracts (contract_id)
                            ON UPDATE CASCADE
                            ON DELETE CASCADE,
                    UNIQUE (contract_id));""")

        cur.execute(
            """CREATE TABLE universe_memberships (
                    membership_id INTEGER NOT NULL PRIMARY KEY,
                    contract_id INTEGER,
                    universe TEXT,
                    FOREIGN KEY (contract_id)
                        REFERENCES contracts (contract_id)
                            ON UPDATE CASCADE
                            ON DELETE CASCADE,
                    UNIQUE (contract_id, universe));""")

        conn.commit()
        cur.close()
        self.disconnect(conn)
        logger.debug("Created database schema.")


class DBNotInitializedError(Exception):
    """Custom exception for connecting to a non-present database."""

    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)



from pathlib import Path
import logging
from shutil import copyfile

logger = logging.getLogger(__name__)


class ConfigInitializer():
    """Initialize the configuration file, if not present"""

    def __init__(self) -> None:
        pass

    @classmethod
    def initalize_config(cls, filepath: Path) -> None:
        """Checks for the presence of a configuration file for the current 
        user. If not present, creates the folders and a default configuration 
        file

        :param filepath: Path to the configuration file
        :type filepath: Path
        """

        cls.__create_directories()
        cls.__create_config_file(
            source_path=Path("default_config.ini"),
            destination_path=filepath)

    @classmethod
    def __create_directories(cls) -> None:
        try:
            directories = (Path.home() / ".barbucket/tv_screener")
            Path.mkdir(directories, parents=True)
        except FileExistsError:
            logger.debug(f"Necessary directories already exist.")
        else:
            logger.info(f"Created directories '{directories}'")

    @classmethod
    def __create_config_file(
            cls,
            source_path: Path,
            destination_path: Path) -> None:
        if Path.is_file(destination_path):
            logger.debug(f"Config file already exists.")
        else:
            copyfile(source_path, destination_path)
            logger.info(
                f"Created config file {destination_path} from default file.")



import logging
from pathlib import Path
from os import path, listdir
from typing import Any, List, Dict

import pandas as pd
import enlighten

from .contracts_db_connector import ContractsDbConnector
from .tv_details_db_connector import TvDetailsDbConnector
from .encodings import Api, Exchange, Symbol


logger = logging.getLogger(__name__)


class TvDetailsProcessor():
    """Processing of contract details provided by Tradingview screener"""

    def __init__(self) -> None:
        self.__contracts_db_connector = ContractsDbConnector()
        self.__tv_details_db_connector = TvDetailsDbConnector()
        self.__file_row: Any = None
        manager = enlighten.get_manager()  # Setup progress bar
        self.__pbar = manager.counter(
            total=0,
            desc="Contracts", unit="contracts")

    def read_tv_data(self) -> None:
        """Read contract details from tv files and write to database"""

        files = self.__get_files_from_dir()
        for file in files:
            logger.info(f"Processing {file}")
            file_data = self.__get_contracts_from_file(file=file)
            self.__pbar.count = 0
            self.__pbar.total = len(file_data)
            n_found = 0
            for row in file_data:
                self.__file_row = row
                try:
                    contract_id = self.__get_contract_id_from_db()
                except TvQueryResultError as e:
                    logger.info(e)
                else:
                    self.__write_contract_details_to_db(contract_id)
                    n_found += 1
                self.__pbar.update(inc=1)
            logger.info(
                f"Added details for {n_found} of {len(file_data)} contracts "
                f"in file '{file}'.")

    def __get_files_from_dir(self) -> List[str]:
        """Create list of paths to all *.csv files in directory"""

        logger.debug(
            "Creating list of paths to all *.csv files in tv-directory.")
        dir_path = Path.home() / ".barbucket/tv_screener"  # Todo: Config
        tv_files = [path.join(dir_path, f) for f in listdir(
            dir_path) if f.endswith(".csv")]  # This also excludes directories
        return tv_files

    def __get_contracts_from_file(self, file: str) -> List[Dict[str, Any]]:
        """Create formatted list of all contracts of a tv file"""

        df = pd.read_csv(file, sep=",")
        file_contracts = []

        # Iterate over file rows
        for _, row in df.iterrows():
            row_formated: Dict[str, Any] = {}

            # Prepare the data
            row_formated['ticker'] = Symbol.decode(
                name=row['Ticker'],
                from_api=Api.TV)
            row_formated['exchange'] = Exchange.decode(
                name=row['Exchange'],
                from_api=Api.TV)
            row_formated['market_cap'] = int(row['Market Capitalization'])
            avg_vol_30_in_curr = row["Average Volume (30 day)"] * \
                row["Simple Moving Average (30)"]
            if pd.isna(avg_vol_30_in_curr):
                row_formated['avg_vol_30_in_curr'] = 0
            else:
                row_formated['avg_vol_30_in_curr'] = int(avg_vol_30_in_curr)
            row_formated['country'] = row['Country']
            if pd.isna(row["Number of Employees"]):
                row_formated['employees'] = 0
            else:
                row_formated['employees'] = int(row["Number of Employees"])
            if pd.isna(row["Gross Profit (FY)"]):
                row_formated['profit'] = 0
            else:
                row_formated['profit'] = int(row["Gross Profit (FY)"])
            if pd.isna(row["Total Revenue (FY)"]):
                row_formated['revenue'] = 0
            else:
                row_formated['revenue'] = int(row["Total Revenue (FY)"])
            file_contracts.append(row_formated)
        return file_contracts

    def __get_contract_id_from_db(self) -> int:
        """Get contract id from db matching some contract info"""

        logger.debug(f"Get contract id from db matching some contract info "
                     "from a tv file row.")
        ticker = self.__file_row['ticker']
        exchange = self.__file_row['exchange']
        filters = {
            'exchange': exchange,
            'contract_type_from_listing': "STOCK",
            'exchange_symbol': ticker}
        return_columns = ['contract_id']
        query_result = self.__contracts_db_connector.get_contracts(
            filters=filters,
            return_columns=return_columns)
        if len(query_result) == 0:
            raise TvQueryResultError(
                f"No contract found in master listing for "
                f"'{self.__file_row['ticker']}' on '"
                f"{self.__file_row['exchange']}'.")
        elif len(query_result) > 1:
            raise TvQueryResultError(
                f"{len(query_result)} contracts found in master listing for '"
                f"{self.__file_row['ticker']}' on '"
                f"{self.__file_row['exchange']}'.")
        else:
            return query_result[0]['contract_id']

    def __write_contract_details_to_db(self, contract_id: int) -> None:
        """Writing tv details to db"""

        self.__tv_details_db_connector.insert_tv_details(
            contract_id=contract_id,
            market_cap=self.__file_row['market_cap'],
            avg_vol_30_in_curr=self.__file_row['avg_vol_30_in_curr'],
            country=self.__file_row['country'],
            employees=self.__file_row['employees'],
            profit=self.__file_row['profit'],
            revenue=self.__file_row['revenue'])


class TvQueryResultError(Exception):
    """Number of contracts matching TV file row is not 1"""

    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)



import signal
import logging
from typing import Any


logger = logging.getLogger(__name__)


class SignalHandler():
    """Handle 'Ctrl+C' commands from the user"""
    __state = False

    def __init__(self) -> None:
        signal.signal(signal.SIGINT, SignalHandler.change_state)

    @classmethod
    def change_state(cls, signum: Any, frame: Any) -> None:
        """Interrupt method, called by system, when Ctrl+C is detected.

        :param signum: [description]
        :type signum: Any
        :param frame: [description]
        :type frame: Any
        """

        logger.warn(
            f": Ctrl-C detected, gracefully stopping operation. Press again to "
            f"stop immediately.")
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        cls.__state = True

    @classmethod
    def is_exit_requested(cls) -> None:
        """Check if the user pressed Ctrl+C.

        :raises ExitSignalDetectedError: User has pressed 'Ctrl+C'
        """

        if cls.__state:
            raise ExitSignalDetectedError("Message")


class ExitSignalDetectedError(Exception):
    """User has pressed 'Ctrl+C'"""

    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)



from enum import Enum


class Api(Enum):
    """Enumeration of available APIs"""
    IB = 1
    TV = 2


class EncodingBase(Enum):
    """Base class for encodings"""

    @classmethod
    def encode(cls, name: str, to_api: Api) -> str:
        """Encode from Barbucket notation to specific API notation

        :param name: Name to encode
        :type name: str
        :param to_api: Api to encode to
        :type to_api: Api
        :raises AttributeError: Name was not found in encoding table
        :return: Encoded name
        :rtype: str
        """

        for element in cls:
            if element.name == name:
                return element.value[to_api]
        raise AttributeError(f"Attribute '{name}' not found.")

    @classmethod
    def decode(cls, name: str, from_api: Api) -> str:
        """Decode from specific API notation to Barbucket notation

        :param name: Name to decode
        :type name: str
        :param from_api: Api to decode from
        :type from_api: Api
        :raises AttributeError: Name was not found in encoding table
        :return: Decoded name
        :rtype: str
        """

        for element in cls:
            if element.value[from_api] == name:
                return element.name
        raise AttributeError(f"Attribute '{name}' not found.")


class Exchange(EncodingBase):
    """Decode and encode exchange names to specific API notations"""

    AEB = {Api.IB: 'AEB', Api.TV: ''}
    AMEX = {Api.IB: 'AMEX', Api.TV: ''}
    ARCA = {Api.IB: 'ARCA', Api.TV: 'NYSE ARCA'}
    ATH = {Api.IB: 'ATH', Api.TV: ''}
    BATS = {Api.IB: 'BATS', Api.TV: ''}
    BM = {Api.IB: 'BM', Api.TV: ''}
    BVL = {Api.IB: 'BVL', Api.TV: ''}
    BVME = {Api.IB: 'BVME', Api.TV: ''}
    BVME_ETF = {Api.IB: 'BVME.ETF', Api.TV: ''}
    EBS = {Api.IB: 'EBS', Api.TV: ''}
    ENEXT_BE = {Api.IB: 'ENEXT.BE', Api.TV: ''}
    FWB = {Api.IB: 'FWB', Api.TV: 'FWB'}
    GETTEX = {Api.IB: 'GETTEX', Api.TV: ''}
    HEX = {Api.IB: 'HEX', Api.TV: ''}
    ISED = {Api.IB: 'ISED', Api.TV: ''}
    LSE = {Api.IB: 'LSE', Api.TV: ''}
    LSE_ETF = {Api.IB: 'LSEETF', Api.TV: ''}
    NASDAQ = {Api.IB: 'ISLAND', Api.TV: 'NASDAQ'}
    NYSE = {Api.IB: 'NYSE', Api.TV: 'NYSE'}
    SBF = {Api.IB: 'SBF', Api.TV: ''}
    SWB = {Api.IB: 'SWB', Api.TV: ''}
    VSE = {Api.IB: 'VSE', Api.TV: ''}
    XETRA = {Api.IB: 'IBIS', Api.TV: 'XETR'}
    XETRA_2 = {Api.IB: 'IBIS2', Api.TV: ''}

    @classmethod
    def decode(cls, name: str, from_api: Api) -> str:
        """Decode from specific api notation to Barbucket notation"""

        # IB inconsistently uses the names 'ISLAND' and 'NASDAQ'
        if (from_api == Api.IB) and (name == 'NASDAQ'):
            name = 'ISLAND'
        return super().decode(name=name, from_api=from_api)


class ContractType(EncodingBase):
    """Decode and encode contract type names to specific API notations"""

    ADR = {Api.IB: 'ADR', Api.TV: ''}
    BOND = {Api.IB: 'BOND', Api.TV: ''}
    CLOSED_END_FUND = {Api.IB: 'CLOSED-END FUND', Api.TV: ''}
    COMMON_STOCK = {Api.IB: 'COMMON', Api.TV: ''}
    CONV_PREFERRED = {Api.IB: 'CONVPREFERRED', Api.TV: ''}
    DUTCH_CERT = {Api.IB: 'DUTCH CERT', Api.TV: ''}
    ETC = {Api.IB: 'ETC', Api.TV: ''}
    ETF = {Api.IB: 'ETF', Api.TV: ''}
    ETN = {Api.IB: 'ETN', Api.TV: ''}
    ETP = {Api.IB: 'ETP', Api.TV: ''}
    FUND_OF_FUNDS = {Api.IB: 'FUND OF FUNDS', Api.TV: ''}
    GDR = {Api.IB: 'GDR', Api.TV: ''}
    GERMAN_CERT = {Api.IB: 'GERMAN CERT', Api.TV: ''}
    LTD_PART = {Api.IB: 'LTD PART', Api.TV: ''}
    MLP = {Api.IB: 'MLP', Api.TV: ''}
    NY_REG_SHRS = {Api.IB: 'NY REG SHRS', Api.TV: ''}
    OPEN_END_FUND = {Api.IB: 'OPEN-END FUND', Api.TV: ''}
    PREFERENCE = {Api.IB: 'PREFERENCE', Api.TV: ''}
    PREFERRED = {Api.IB: 'PREFERRED', Api.TV: ''}
    REIT = {Api.IB: 'REIT', Api.TV: ''}
    RIGHT = {Api.IB: 'RIGHT', Api.TV: ''}
    ROYALTY_TRUST = {Api.IB: 'ROYALTY TRST', Api.TV: ''}
    SAVINGS_SHARE = {Api.IB: 'SAVINGS SHARE', Api.TV: ''}
    TRACKING_STOCK = {Api.IB: 'TRACKING STK', Api.TV: ''}
    UNIT = {Api.IB: 'UNIT', Api.TV: ''}
    US_DOMESTIC = {Api.IB: 'US DOMESTIC', Api.TV: ''}


class Symbol():
    """Decode and encode symbol names to specific API notations"""

    @classmethod
    def encode(cls, name: str, to_api: Api) -> str:
        """Encode from Barbucket notation to specific API notation"""

        if to_api == Api.IB:
            name = name.replace("_", " ")
        elif to_api == Api.TV:
            name = name.replace("_", ".")
        return name

    @classmethod
    def decode(cls, name: str, from_api: Api) -> str:
        """Decode from specific API notation to Barbucket notation"""

        if from_api == Api.IB:
            name = name.replace(" ", "_")
        elif from_api == Api.TV:
            name = name.replace(".", "_")
        return name



import logging
from typing import Any, List

import enlighten
from ib_insync.wrapper import RequestError

from .signal_handler import SignalHandler, ExitSignalDetectedError
from .contracts_db_connector import ContractsDbConnector
from .ib_details_db_connector import IbDetailsDbConnector
from .tws_connector import TwsConnector, IbDetailsInvalidError

logger = logging.getLogger(__name__)


class IbDetailsProcessor():
    """Downloading of contract details from IB TWS and storing to db"""

    def __init__(self) -> None:
        self.__contracts_db_connector = ContractsDbConnector()
        self.__ib_details_db_connector = IbDetailsDbConnector()
        self.__tws_connector = TwsConnector()
        self.__contracts: List[Any] = []
        self.__details: Any = None
        self.__pbar: Any = None
        self.__signal_handler = SignalHandler()
        manager = enlighten.get_manager()  # Setup progress bar
        self.__pbar = manager.counter(
            total=0,
            desc="Contracts", unit="contracts")

    def update_ib_contract_details(self) -> None:
        """Download and store all missing contract details entries from IB TWS.
        """

        self.__get_contracts()
        self.__pbar.total = len(self.__contracts)
        self.__tws_connector.connect()
        try:
            for contract in self.__contracts:
                self.__signal_handler.is_exit_requested()
                logger.info(
                    f"{contract['broker_symbol']}_{contract['exchange']}_"
                    f"{contract['currency']}")
                try:
                    self.__get_contract_details_from_tws(contract)
                except RequestError as e:
                    logger.info(e)
                    continue
                except IbDetailsInvalidError as e:
                    logger.info(e)
                    continue
                else:
                    self.__insert_ib_details_into_db(contract)
                finally:
                    self.__pbar.update(inc=1)
        except ExitSignalDetectedError:
            pass
        else:
            logger.info(
                f"Updated IB details for master listings.")
        finally:
            self.__tws_connector.disconnect()

    def __get_contracts(self) -> None:
        """Get contracts from db, where IB details are missing"""

        return_columns = [
            'contract_id', 'broker_symbol', 'exchange', 'currency']
        filters = {'primary_exchange': "NULL"}
        self.__contracts = self.__contracts_db_connector.get_contracts(
            filters=filters,
            return_columns=return_columns)
        logger.info(f"Found {len(self.__contracts)} contracts with missing "
                    f"IB details in master listing.")

    def __get_contract_details_from_tws(self, contract: Any) -> None:
        self.__details = self.__tws_connector.download_contract_details(
            broker_symbol=contract['broker_symbol'],
            exchange=contract['exchange'],
            currency=contract['currency'])

    def __insert_ib_details_into_db(self, contract: Any) -> None:
        self.__ib_details_db_connector.insert_ib_details(
            contract_id=contract['contract_id'],
            contract_type_from_details=self.__details.stockType,
            primary_exchange=self.__details.contract.primaryExchange,
            industry=self.__details.industry,
            category=self.__details.category,
            subcategory=self.__details.subcategory)



import logging
from typing import Any, List

from .contracts_db_connector import ContractsDbConnector
from .ib_exchange_listing_reader import (
    IbExchangeListingSinglepageReader,
    IbExchangeListingMultipageReader,
    WebscrapingReturnedNoResultError)
from .signal_handler import ExitSignalDetectedError


logger = logging.getLogger(__name__)


class IbExchangeListingsProcessor():
    """Provides methods to sync local exchange listings to the IB website."""

    def __init__(self) -> None:
        self.__contracts_db_connector = ContractsDbConnector()
        self.__ctype: str = ""
        self.__exchange: str = ""
        self.__website_contracts: List[Any] = []
        self.__database_contracts: List[Any] = []

    def sync_contracts_to_listing(self, ctype: str, exchange: str) -> None:
        """Sync local exchange listings to the IB website

        :param ctype: Contracts type to sync
        :type ctype: str
        :param exchange: Exchange to sync to
        :type exchange: str
        """

        self.__ctype = ctype
        self.__exchange = exchange

        pagetypes = {
            'STOCK': "paginated",
            'ETF': "single"
        }

        if pagetypes[ctype] == "single":
            scraper = IbExchangeListingSinglepageReader()
        elif pagetypes[ctype] == "paginated":
            scraper = IbExchangeListingMultipageReader()
        try:
            self.__website_contracts = scraper.read_ib_exchange_listing(
                self.__ctype,
                self.__exchange)
            self.__get_contracts_from_db()
            removed_count = self.__remove_deleted_contracts_from_db()
            added_count = self.__add_new_contracts_to_db()
        except ExitSignalDetectedError:
            logger.info("Stopped.")
        except WebscrapingReturnedNoResultError:
            logger.error(f"Webscraping for {self.__ctype} on "
                         f"{self.__exchange} returned no results.")
        else:
            logger.info(
                f"Master listing synced for '{ctype}' on '{exchange}'. "
                f"{added_count} contracts were added, {removed_count} were "
                f"removed.")

    def __get_contracts_from_db(self) -> None:
        filters = {
            'contract_type_from_listing': self.__ctype,
            'exchange': self.__exchange}
        return_columns = ['broker_symbol', 'currency']
        self.__database_contracts = self.__contracts_db_connector.get_contracts(
            filters=filters,
            return_columns=return_columns)

    def __remove_deleted_contracts_from_db(self) -> int:
        removed_count = 0
        for db_row in self.__database_contracts:
            exists = False
            for web_row in self.__website_contracts:
                if ((db_row['broker_symbol'] == web_row['broker_symbol'])
                        and (db_row['currency'] == web_row['currency'])):
                    exists = True
                    break
            if not exists:
                self.__contracts_db_connector.delete_contract(
                    symbol=db_row['broker_symbol'],
                    exchange=self.__exchange,
                    currency=db_row['currency'])
                removed_count += 1
        return removed_count

    def __add_new_contracts_to_db(self) -> int:
        added_count = 0
        for web_row in self.__website_contracts:
            exists = False
            for db_row in self.__database_contracts:
                if ((web_row['broker_symbol'] == db_row['broker_symbol'])
                        and (web_row['currency'] == db_row['currency'])):
                    exists = True
                    break
            if not exists:
                self.__contracts_db_connector.create_contract(
                    contract_type_from_listing=self.__ctype,
                    exchange_symbol=web_row['exchange_symbol'],
                    broker_symbol=web_row['broker_symbol'],
                    name=web_row['name'],
                    currency=web_row['currency'],
                    exchange=self.__exchange)
                added_count += 1
        return added_count



import sqlite3
import logging
from typing import Dict, List

from .db_connector import DbConnector

logger = logging.getLogger(__name__)


class ContractsDbConnector(DbConnector):
    """Provides methods to access the 'contracts' table of the database."""

    def __init__(self) -> None:
        super().__init__()

    def create_contract(self, contract_type_from_listing: str,
                        exchange_symbol: str, broker_symbol: str, name: str,
                        currency: str, exchange: str) -> None:
        """Creates a new contract in the database

        :param contract_type_from_listing: Type of contract from the listing
        :type contract_type_from_listing: str
        :param exchange_symbol: Contract symbol at exchange
        :type exchange_symbol: str
        :param broker_symbol: Contract symbol at broker
        :type broker_symbol: str
        :param name: Contract name
        :type name: str
        :param currency: Contract currency
        :type currency: str
        :param exchange: Contract exchange
        :type exchange: str
        """
        # Todo: what if already exits?

        conn = self.connect()
        cur = conn.cursor()
        cur.execute(
            """INSERT INTO contracts (
                    contract_type_from_listing,
                    exchange_symbol,
                    broker_symbol,
                    name,
                    currency,
                    exchange)
                    VALUES (?, ?, ?, ?, ?, ?)""",
            (contract_type_from_listing,
                exchange_symbol,
                broker_symbol,
                name,
                currency,
                exchange))
        conn.commit()
        cur.close()
        self.disconnect(conn)
        logger.debug(
            f"Created new contract_db entry: {contract_type_from_listing}_"
            f"{exchange}_{broker_symbol}_{currency}.")

    def get_contracts(self, filters: Dict = {}, return_columns: List = []
                      ) -> List[sqlite3.Row]:
        """Returns the requsted columns of the contracts table, filtered by 
        the given filters

        :param filters: Equality conditions for filtering, formatted as a 
        dictionary: {'column_name': "value"}. Value can be int, float or 
        string, defaults to {}
        :type filters: Dict, optional
        :param return_columns: List of column names to return, empty list 
        returns all columns, defaults to []
        :type return_columns: List, optional
        :return: List of rows matching the given parameters
        :rtype: List[sqlite3.Row]
        """

        # Prepare query
        query = "SELECT * FROM all_contract_info"
        if len(return_columns) > 0:
            cols = ", ".join(return_columns)
            query = query.replace("*", cols)
        if len(filters) > 0:
            query += " WHERE "
            for key, value in filters.items():
                if value == "NULL":
                    query += (key + " IS " + str(value) + " and ")
                elif isinstance(value, str):
                    query += (key + " = '" + str(value) + "' and ")
                elif isinstance(value, (int, float)):
                    query += (key + " = " + str(value) + " and ")
            query = query[:-5]  # remove trailing 'and'
            query += ";"

        # Get requested values from db
        conn = self.connect()
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        cur.execute(query)
        contracts = cur.fetchall()
        conn.commit()
        cur.close()
        self.disconnect(conn)
        logger.debug(f"Qurey '{query}' returned {len(contracts)} results.")
        return contracts

    def delete_contract(self, exchange: str, symbol: str, currency: str) -> None:
        """Deletets a contract from the db.

        :param exchange: Exchange of the contract to delete
        :type exchange: str
        :param symbol: Symbol of the contract to delete
        :type symbol: str
        :param currency: Currency of the contract to delete
        :type currency: str
        """

        conn = self.connect()
        cur = conn.cursor()
        cur.execute(
            """DELETE FROM contracts
                    WHERE (broker_symbol = ?
                        AND exchange = ?
                        AND currency = ?);""",
            (symbol, exchange, currency))
        conn.commit()
        cur.close()
        self.disconnect(conn)
        logger.debug(f"Deleted contract: {exchange}_{symbol}_{currency}")

    def delete_contract_id(self, contract_id: int) -> None:
        """Deletets a contract from the db.

        :param contract_id: Contract ID of the contract to delete
        :type contract_id: int
        """

        conn = self.connect()
        cur = conn.cursor()
        cur.execute(
            """DELETE FROM contracts
                    WHERE contract_id = ?;""",
            (contract_id,))
        conn.commit()
        cur.close()
        self.disconnect(conn)
        logger.debug(f"Deleted contract with id {contract_id}")



import configparser
from pathlib import Path
from typing import List

from .config_initializer import ConfigInitializer


class ConfigReader():
    """Read config values from existing configuration file."""

    __parser = configparser.ConfigParser(allow_no_value=True)
    __CONFIG_FILE_PATH = Path.home() / ".barbucket/config.ini"
    __is_initialized = False

    def __init__(self) -> None:
        if not ConfigReader.__is_initialized:
            self.__initializer = ConfigInitializer()
            self.__initializer.initalize_config(
                ConfigReader.__CONFIG_FILE_PATH)
            ConfigReader.__is_initialized = True
        ConfigReader.__parser.read(self.__CONFIG_FILE_PATH)

    def get_config_value_single(self, section: str, option: str) -> str:
        """Reads a single config value from the config file

        :param section: Section of the config file
        :type section: str
        :param option: Option of the config file
        :type option: str
        :return: Value from the config file
        :rtype: str
        """
        return ConfigReader.__parser.get(section, option)

    def get_config_value_list(self, section: str, option: str) -> List[str]:
        """Reads a config value list from the config file

        :param section: Section of the config file
        :type section: str
        :param option: Option of the config file
        :type option: str
        :return: Values from the config file
        :rtype: List[str]
        """
        value = ConfigReader.__parser.get(section, option)
        return value.split(",")  # split by comma and change to list



import logging

from .db_connector import DbConnector

logger = logging.getLogger(__name__)


class TvDetailsDbConnector(DbConnector):
    """Provides methods to access the 'quotes' table of the database."""

    def __init__(self) -> None:
        super().__init__()

    def insert_tv_details(
            self, contract_id: int, market_cap: int, avg_vol_30_in_curr: int,
            country: str, employees: int, profit: int, revenue: int) -> None:
        """Writing tv details to db

        :param contract_id: Contract ID to insert details for
        :type contract_id: int
        :param market_cap: Market capitalization
        :type market_cap: int
        :param avg_vol_30_in_curr: Average trading volume of last 30 days in 
        contract's currency
        :type avg_vol_30_in_curr: int
        :param country: Country of origin
        :type country: str
        :param employees: Number of employees
        :type employees: int
        :param profit: Last annual profit
        :type profit: int
        :param revenue: Last annual revenue
        :type revenue: int
        """

        conn = self.connect()
        cur = conn.cursor()
        cur.execute("""
            REPLACE INTO contract_details_tv (
                contract_id,
                market_cap,
                avg_vol_30_in_curr,
                country,
                employees,
                profit,
                revenue)
            VALUES (?, ?, ?, ?, ?, ?, ?)
            """, (
            contract_id,
            market_cap,
            avg_vol_30_in_curr,
            country,
            employees,
            profit,
            revenue))
        conn.commit()
        cur.close()
        self.disconnect(conn)
        logger.debug(f"Wrote tv details for contract_id {contract_id} to db.")



from typing import List, Any
import logging

import pandas as pd

from .db_connector import DbConnector

logger = logging.getLogger(__name__)


class QuotesDbConnector(DbConnector):
    """Provides methods to access the 'quotes' table of the database."""

    def __init__(self) -> None:
        super().__init__()

    def insert_quotes(self, quotes: List[Any]) -> None:
        """Insert quotes into the db

        :param quotes: quotes
        :type quotes: List[Any]
        """

        conn = self.connect()
        cur = conn.cursor()
        cur.executemany("""REPLACE INTO quotes (contract_id, date, open, high, 
            low, close, volume) VALUES (?, ?, ?, ?, ?, ?, ?)""", quotes)
        # Important to 'REPLACE' because last quote of download is incomplete,
        # if quote interval was unfinished. Needs to be replaced by complete
        # quote with overlapping subsequent quotes download
        conn.commit()
        cur.close()
        self.disconnect(conn)
        logger.debug(f"Inserted {len(quotes)} for contract_id {quotes[0][0]} "
                     f"into db.")

    def get_quotes(self, contract_id: int) -> pd.DataFrame:
        """Get quotes from the db

        :param contract_id: Contract ID to supply quotes for
        :type contract_id: int
        :return: All quotes available from the database for the given contract
        :rtype: pd.DataFrame
        """

        query = f"""SELECT date, open, high, low, close, volume
                    FROM quotes
                    WHERE contract_id = {contract_id}
                    ORDER BY date ASC;"""
        conn = self.connect()
        df = pd.read_sql_query(query, conn)
        self.disconnect(conn)
        df['date'] = pd.to_datetime(df['date'], format='%Y-%m-%d')
        df = df.set_index('date')
        logger.debug(f"Read {len(df)} quotes for contract id {contract_id} "
                     f"from db.")
        return df



import sqlite3
import logging
from typing import List

from .db_connector import DbConnector

logger = logging.getLogger(__name__)


class UniversesDbConnector(DbConnector):

    def __init__(self) -> None:
        super().__init__()

    def create_universe(self, name: str, contract_ids: List[int]) -> None:
        """Create a new universe with given contracts

        :param name: New universes name
        :type name: str
        :param contract_ids: Contract IDs of universes members
        :type contract_ids: List[int]
        """

        for contract_id in contract_ids:
            self.__create_membership(contract_id, name)
        logger.debug(
            f"Created universe '{name}' with {len(contract_ids)} contracts.")

    def __create_membership(self, contract_id: int, universe: str) -> None:
        # Todo: What if member does not exist?

        conn = self.connect()
        cur = conn.cursor()
        cur.execute(
            """INSERT INTO universe_memberships (contract_id, universe) 
                VALUES (?, ?)""",
            (contract_id, universe))
        conn.commit()
        cur.close()
        self.disconnect(conn)
        logger.debug(
            f"Created universe membership for {contract_id} in universe "
            f"'{universe}'")

    def get_universes(self) -> List[str]:
        """Get all existing universes

        :return: All existing universes
        :rtype: List[str]
        """

        conn = self.connect()
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        cur.execute("SELECT DISTINCT universe FROM universe_memberships;")
        row_list = cur.fetchall()
        conn.commit()
        cur.close()
        self.disconnect(conn)
        logger.debug(f"Query for universes returned {len(row_list)} entries")
        return [row['universe'] for row in row_list]

    def get_universe_members(self, universe: str) -> List[int]:
        """Get all members of a given universe

        :param universe: Name of the universe
        :type universe: str
        :return: Members of the universe
        :rtype: List[int]
        """

        conn = self.connect()
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        cur.execute(
            """SELECT contract_id 
                FROM universe_memberships 
                WHERE universe = ?;""",
            (universe,))  # Comma, to not iterate over string characters
        row_list = cur.fetchall()
        conn.commit()
        cur.close()
        self.disconnect(conn)
        logger.debug(
            f"Query for members of universe '{universe}' returned {len(row_list)} "
            f"entries")
        return [row['contract_id'] for row in row_list]

    def delete_universe(self, universe: str) -> int:
        """Delete an existing universe

        :param universe: Name of the universe
        :type universe: str
        :return: Number of members
        :rtype: int
        """

        conn = self.connect()
        cur = conn.cursor()
        cur.execute(
            "DELETE FROM universe_memberships WHERE universe = ?;",
            (universe,))  # tuple, to not iterate over string characters
        cur.execute(
            "select changes()")
        n_affeced_rows = cur.fetchone()[0]
        conn.commit()
        cur.close()
        self.disconnect(conn)
        logger.debug(
            f"Deleting universe '{universe}' resulted in {n_affeced_rows} "
            f"affected rows")
        return n_affeced_rows



import time
import logging
from typing import Any, Dict, List
from abc import ABC, abstractmethod

from bs4 import BeautifulSoup
import requests
import enlighten

from .signal_handler import SignalHandler
from .encodings import Api, Exchange, Symbol

logger = logging.getLogger(__name__)


class IbExchangeListingReader(ABC):
    """Abstract baseclass for exchange listings readers"""

    @abstractmethod
    def read_ib_exchange_listing(self, ctype: str, exchange: str) -> Any:
        raise NotImplementedError


class IbExchangeListingSinglepageReader(IbExchangeListingReader):
    """Exchange listings reader for singlepage listings"""

    def __init__(self) -> None:
        self.__html: str = ""
        self.__ctype: str = ""
        self.__exchange: str = ""

    def read_ib_exchange_listing(
            self,
            ctype: str,
            exchange: str) -> List[Dict[str, Any]]:
        """Read contracts from exchange listing website

        :param ctype: Contracts type to read
        :type ctype: str
        :param exchange: Exchange to read from
        :type exchange: str
        :return: Contracts from website
        :rtype: List[Dict[str, Any]]
        """

        self.__ctype = ctype
        self.__exchange = Exchange.encode(name=exchange, to_api=Api.IB)
        self.__get_html()
        self.__correct_ib_error()
        contracts = self.__extract_data()
        self.__validate_result(contracts)
        return contracts

    def __get_html(self) -> None:
        url = (
            f"https://www.interactivebrokers.com/en/index.php?f=567&exch="
            f"{self.__exchange}")
        self.__html = requests.get(url).text

    def __correct_ib_error(self) -> None:
        """Website HTML contains structural errors that prevent parsing."""

        old_lines = self.__html.splitlines()
        new_lines = []
        corrections = 0
        for line in old_lines:
            if (('        <td align="left" valign="middle">' in line)
                    and ("href" not in line)):
                line = line.replace("</a>", "")
                corrections += 1
            new_lines.append(line)
        self.__html = "".join(new_lines)
        if corrections == 0:
            logger.debug(
                f"IB error for singlepage listings no longer present. Checked "
                f"{len(old_lines)} lines.")

    def __extract_data(self) -> List[Dict[str, Any]]:
        soup = BeautifulSoup(self.__html, 'html.parser')
        tables = soup.find_all(
            'table',
            class_='table table-striped table-bordered')

        rows = tables[2].tbody.find_all('tr')
        website_contracts = []
        for row in rows:
            columns = row.find_all('td')
            row_dict = {
                'type': self.__ctype,
                'broker_symbol': Symbol.decode(
                    name=columns[0].text.strip(),
                    from_api=Api.IB),
                'name': columns[1].text.strip(),
                'exchange_symbol': Symbol.decode(
                    name=columns[2].text.strip(),
                    from_api=Api.IB),
                'currency': columns[3].text.strip(),
                'exchange': Exchange.decode(
                    name=self.__exchange,
                    from_api=Api.IB)}
            website_contracts.append(row_dict)
        # todo: log amount
        return website_contracts

    def __validate_result(self, contracts):
        if len(contracts) == 0:
            raise WebscrapingReturnedNoResultError


class IbExchangeListingMultipageReader(IbExchangeListingReader):
    """Exchange listings reader for multipage listings"""

    def __init__(self) -> None:
        self.__signal_handler = SignalHandler()
        self.__current_page = 1
        self.__page_count = 1
        self.__html: str = ""
        self.__website_data: Any = []
        self.__ctype: str = ""
        self.__exchange: str = ""
        manager = enlighten.get_manager()
        self.__pbar = manager.counter(
            total=0,
            desc="Pages", unit="pages")

    def read_ib_exchange_listing(
            self,
            ctype: str,
            exchange: str) -> List[Dict[str, Any]]:
        """Read contracts from exchange listing website

        :param ctype: Contracts type to read
        :type ctype: str
        :param exchange: Exchange to read from
        :type exchange: str
        :return: Contracts from website
        :rtype: List[Dict[str, Any]]
        """

        self.__ctype = ctype
        self.__exchange = Exchange.encode(name=exchange, to_api=Api.IB)

        while self.__current_page <= self.__page_count:
            self.__get_html()
            if self.__current_page == 1:
                self.__set_page_count()
            self.__correct_ib_error()
            self.__extract_data()
            logger.debug(
                f"Scraped IB exchange listing for {self.__exchange}, page "
                f"{self.__current_page}.")
            self.__signal_handler.is_exit_requested()  # raises ex to exit
            self.__pbar.update(incr=1)
            self.__current_page += 1
            if self.__current_page != self.__page_count:
                time.sleep(3)  # show some mercy to IB webserver
        return self.__website_data

    def __set_page_count(self) -> None:
        soup = BeautifulSoup(self.__html, 'html.parser')
        pagination_tables = soup.find_all('ul', class_='pagination')
        page_buttons = pagination_tables[0].find_all('li')
        self.__page_count = int(page_buttons[-2].text)
        self.__pbar.total = self.__page_count
        # todo: handle errors

    def __get_html(self) -> None:
        url = (f"https://www.interactivebrokers.com/en/index.php?f=2222"
               f"&exch={self.__exchange}&showcategories=STK&p=&cc=&limit="
               f"100&page={self.__current_page}")
        self.__html = requests.get(url).text

    def __correct_ib_error(self) -> None:
        """Website HTML contains structural errors that prevent parsing."""

        if ("(click link for more details)</span></th>\n                       </th>"
                in self.__html):
            self.__html = self.__html.replace(
                "(click link for more details)</span></th>\n                       </th>\n",
                "(click link for more details)</span></th>\n")
        else:
            logger.debug(
                f"IB error for paginated listings no longer present. Checked "
                f"{len(self.__html.splitlines())} lines.")

    def __extract_data(self) -> None:
        soup = BeautifulSoup(self.__html, 'html.parser')
        tables = soup.find_all(
            'table',
            class_='table table-striped table-bordered')
        rows = tables[2].tbody.find_all('tr')

        for row in rows:
            cols = row.find_all('td')
            row_dict = {
                'type': self.__ctype,
                'broker_symbol': Symbol.decode(
                    name=cols[0].text.strip(),
                    from_api=Api.IB),
                'name': cols[1].text.strip(),
                'exchange_symbol': Symbol.decode(
                    name=cols[2].text.strip(),
                    from_api=Api.IB),
                'currency': cols[3].text.strip(),
                'exchange': Exchange.decode(
                    name=self.__exchange,
                    from_api=Api.IB)}
            self.__website_data.append(row_dict)
        # todo: log ammount
        # todo: handle ammount == 0


class WebscrapingReturnedNoResultError(Exception):
    """Obviously something went wrong."""

    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)



import logging
import logging.handlers
from pathlib import Path

import click

from .universes_db_connector import UniversesDbConnector
from .ib_exchange_listings_processor import IbExchangeListingsProcessor
from .ib_details_processor import IbDetailsProcessor
from .tv_details_processor import TvDetailsProcessor
from .ib_quotes_processor import IbQuotesProcessor
from .encodings import Exchange


logger = logging.getLogger(__name__)


@click.group()
# Initial group. This method is called to run the cli.
@click.option("-v", "--verbose", count=True, default=0, help="-v for DEBUG")
def cli(verbose) -> None:
    if verbose > 0:  # -v -> 1, -vv -> 2, etc.
        root_logger = logging.getLogger()
        root_logger.handlers[0].level = logging.DEBUG  # Dirty


# Group contracts
@cli.group()
def contracts() -> None:
    """Contracts commands"""


@contracts.command()
@click.option("-t", "--type", "contract_type", required=True, type=str)
@click.option("-e", "--exchange", "exchange", required=True, type=str)
def sync_listing(contract_type: str, exchange: str) -> None:
    """Sync master listing to IB exchange listing"""
    logger.debug(
        f"User requested to sync '{contract_type}' contracts on '{exchange}' "
        f"to master listing via the cli.")
    if validate_exchange(exchange.upper()):
        ib_exchange_listings_processor = IbExchangeListingsProcessor()
        ib_exchange_listings_processor.sync_contracts_to_listing(
            ctype=contract_type.upper(),
            exchange=exchange.upper()
        )


@contracts.command()
def download_ib_details() -> None:
    """Fetch details for all contracts from IB TWS"""
    logger.debug(
        f"User requested to download details from TWS via the cli.")
    ib_details_processor = IbDetailsProcessor()
    ib_details_processor.update_ib_contract_details()


@contracts.command()
def read_tv_details() -> None:
    """Read details for all contracts from TV files"""

    logger.debug(
        f"User requested to read and store details from tv files via the cli.")
    tv_details_processor = TvDetailsProcessor()
    tv_details_processor.read_tv_data()


# Group quotes
@cli.group()
def quotes() -> None:
    """Quotes commands"""


@quotes.command()
@click.option("-u", "--universe", "universe", required=True, type=str)
def fetch(universe: str) -> None:
    """Fetch quotes from IB TWS"""

    universe = universe.upper()
    logger.debug(
        f"User requested to download quotes from TWS for universe "
        f"'{universe}' via the cli.")
    ib_quotes_processor = IbQuotesProcessor()
    ib_quotes_processor.download_historical_quotes(
        universe=universe)


# Group universes
@cli.group()
def universes() -> None:
    """Universes commands"""


@universes.command()
@click.option("-n", "--name", "name", required=True, type=str)
@click.option("-c", "--contract_ids", "contract_ids", required=True, type=str)
def create(name: str, contract_ids: str) -> None:
    """Create new universe"""

    name = name.upper()
    logger.debug(
        f"User requested to create universe '{name}' with {len(contract_ids)} "
        f"members via the cli.")
    con_list = [int(n) for n in contract_ids.split(",")]
    universes_db_connector = UniversesDbConnector()
    universes_db_connector.create_universe(
        name=name,
        contract_ids=con_list)
    logger.info(f"Created universe '{name}' with {len(con_list)} contracts.")


@universes.command()
def list() -> None:
    """List all universes"""

    logger.debug(
        f"User requested to list all existing universes via the cli.")
    universes_db_connector = UniversesDbConnector()
    universes = universes_db_connector.get_universes()
    if universes:
        logger.info(f"Existing universes: {universes}")
    else:
        logger.info(f"No existing universes")


@universes.command()
@click.option("-n", "--name", "name", required=True, type=str)
def members(name: str) -> None:
    """List universes members"""

    name = name.upper()
    logger.debug(
        f"User requested to list the members of universe '{name}' via the cli.")
    universes_db_connector = UniversesDbConnector()
    members = universes_db_connector.get_universe_members(universe=name)
    if members:
        logger.info(f"Members of universe '{name}': {members}")
    else:
        logger.info(f"Universe '{name}' does not exist.")


@universes.command()
@click.option("-n", "--name", "name", required=True, type=str)
@click.confirmation_option(
    prompt=("Are you sure you want to delete this universe?"))
def delete(name: str) -> None:
    """Delete universe"""

    name = name.upper()
    logger.debug(
        f"User requested to delete the universe '{name}' via the cli.")
    universes_db_connector = UniversesDbConnector()
    n_affeced_rows = universes_db_connector.delete_universe(universe=name)
    if n_affeced_rows > 0:
        logger.info(
            f"Deleted universe '{name}' with {n_affeced_rows} members.")
    else:
        logger.info(f"Universe '{name}' did not exist.")


def validate_exchange(exchange) -> bool:
    exchanges = [member.name for member in Exchange]
    if not exchange in exchanges:
        logger.info(f"'{exchange}' is not a valid exchange. Valid exchanges "
                    f"are: {exchanges}")
        return False
    return True



import logging

from .db_connector import DbConnector

logger = logging.getLogger(__name__)


class IbDetailsDbConnector(DbConnector):
    """Provides methods to access the 'ib_details' table of the database."""

    def __init__(self) -> None:
        super().__init__()

    def insert_ib_details(self, contract_id: str,
                          contract_type_from_details: str,
                          primary_exchange: str, industry: str, category: str,
                          subcategory: str) -> None:
        """Insert contract details into db

        :param contract_id: Contract ID of the contract
        :type contract_id: str
        :param contract_type_from_details: Type of the contract
        :type contract_type_from_details: str
        :param primary_exchange: Primary exchange of the contract
        :type primary_exchange: str
        :param industry: Industry classification of the contract
        :type industry: str
        :param category: Category classification of the contract
        :type category: str
        :param subcategory: Subcategory classification of the contract
        :type subcategory: str
        """

        conn = self.connect()
        cur = conn.cursor()
        cur.execute("""
            REPLACE INTO contract_details_ib (
                contract_id,
                contract_type_from_details,
                primary_exchange,
                industry,
                category,
                subcategory)
                VALUES (?, ?, ?, ?, ?, ?)""", (
                    contract_id,
                    contract_type_from_details,
                    primary_exchange,
                    industry,
                    category,
                    subcategory))
        conn.commit()
        cur.close()
        self.disconnect(conn)
        logger.debug(f"Inserted IB details into db: {contract_id} "
                     f"{contract_type_from_details} {primary_exchange} "
                     f"{industry} {category} {subcategory}")



import sqlite3
from typing import Any
import logging

from .db_connector import DbConnector

logger = logging.getLogger(__name__)


class QuotesStatusDbConnector(DbConnector):
    """Provides methods to access the 'quotes_status' table of the database.

    Status code:
    0: No quotes downloaded yet
    1: Successfully downloaded quotes
    >1: TWS error code
    """

    def __init__(self) -> None:
        super().__init__()

    def get_quotes_status(self, contract_id: int) -> Any:
        """Get a contract's quote status from the db. Create, if entry does not 
        exist.

        :param contract_id: Contract ID to provide the quotes download status 
        for
        :type contract_id: int
        :return: Current quotes download status from the database for the 
        given contract
        :rtype: Any
        """

        if not self.__check_status_exists(contract_id=contract_id):
            self.__create_status_entry(contract_id=contract_id)
        return self.__get_status(contract_id=contract_id)

    def __check_status_exists(self, contract_id: int) -> Any:
        conn = self.connect()
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        cur.execute(
            """SELECT COUNT(*) FROM quotes_status WHERE contract_id = ?;""",
            (contract_id,))
        count = cur.fetchone()
        conn.commit()
        cur.close()
        self.disconnect(conn)
        if count[0] > 0:
            return True
        else:
            return False

    def __get_status(self, contract_id: int) -> Any:
        conn = self.connect()
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        cur.execute("""SELECT *
                    FROM quotes_status
                    WHERE contract_id = ?;""",
                    (contract_id,))
        result = cur.fetchone()
        conn.commit()
        cur.close()
        self.disconnect(conn)
        return result

    def __create_status_entry(self, contract_id: int) -> None:
        conn = self.connect()
        cur = conn.cursor()
        cur.execute(
            """REPLACE into quotes_status(
                contract_id,
                status_code,
                status_text,
                daily_quotes_requested_from,
                daily_quotes_requested_till) VALUES(?, ?, ?, ?, ?)""",
            (contract_id,
                0,
                "No quotes downloaded yet.",
                "NULL",
                "NULL"))
        conn.commit()
        cur.close()
        self.disconnect(conn)
        logger.debug(f"Created new quotes_status_db entry: {contract_id}.")

    def update_quotes_status(self, contract_id: int, status_code: int,
                             status_text: str,
                             daily_quotes_requested_from: str,
                             daily_quotes_requested_till: str) -> None:
        """Update a contract's quote status in the db.

        :param contract_id: Contract ID to update the quotes download status 
        for
        :type contract_id: int
        :param status_code: Status code from TWS
        :type status_code: int
        :param status_text: Status text, from TWS if error
        :type status_text: str
        :param daily_quotes_requested_from: Earlist date that quotes were 
        requested for
        :type daily_quotes_requested_from: str
        :param daily_quotes_requested_till: Latest date that quotes were 
        requested for
        :type daily_quotes_requested_till: str
        """

        conn = self.connect()
        cur = conn.cursor()
        cur.execute(
            """REPLACE into quotes_status(
                contract_id,
                status_code,
                status_text,
                daily_quotes_requested_from,
                daily_quotes_requested_till) VALUES(?, ?, ?, ?, ?)""",
            (contract_id,
             status_code,
             status_text,
             daily_quotes_requested_from,
             daily_quotes_requested_till))
        conn.commit()
        cur.close()
        self.disconnect(conn)
        logger.debug(f"Inserted quotes status into db: '{contract_id}', "
                     f"'{status_code}', '{status_text}', "
                     f"'{daily_quotes_requested_from}', "
                     f"'{daily_quotes_requested_till}'")



import logging
from typing import Any, List
from datetime import date, timedelta
from sqlite3 import Row

import numpy as np
import enlighten
from ib_insync.wrapper import RequestError
from ib_insync.objects import BarDataList

from .signal_handler import SignalHandler, ExitSignalDetectedError
from .config_reader import ConfigReader
from .contracts_db_connector import ContractsDbConnector
from .universes_db_connector import UniversesDbConnector
from .quotes_db_connector import QuotesDbConnector
from .quotes_status_db_connector import QuotesStatusDbConnector
from .tws_connector import TwsConnector


logger = logging.getLogger(__name__)


class IbQuotesProcessor():
    """Provides methods to download historical quotes from TWS."""

    def __init__(self) -> None:
        self.__contracts_db_connector = ContractsDbConnector()
        self.__universes_db_connector = UniversesDbConnector()
        self.__quotes_db_connector = QuotesDbConnector()
        self.__quotes_status_db_connector = QuotesStatusDbConnector()
        self.__tws_connector = TwsConnector()
        self.__signal_handler = SignalHandler()
        self.__config_reader = ConfigReader()
        self.__contract_id: int
        self.__contract_data: Row
        self.__quotes_status: Row
        self.__duration: str
        self.__quotes_from: str
        self.__quotes_till: str
        manager = enlighten.get_manager()  # Setup progress bar
        self.__pbar = manager.counter(
            total=0,
            desc="Contracts", unit="contracts")

    def download_historical_quotes(self, universe: str) -> None:
        """Download historical quotes from TWS

        :param universe: Universe to download quotes for
        :type universe: str
        """

        contract_ids = self.__get_contracts(universe)
        self.__pbar.total = len(contract_ids)
        self.__tws_connector.connect()
        try:
            for self.__contract_id in contract_ids:
                self.__signal_handler.is_exit_requested()
                self.__get_contract_data()
                self.__get_contract_status()
                try:
                    self.__calculate_dates()
                except (QuotesDurationError, ContractHasErrorStatusError):
                    # Decrement total, as these contracts add almost nothing to time
                    self.__pbar.total -= 1
                    self.__pbar.update(incr=0)
                    continue
                try:
                    bar_data = self.__get_quotes_from_tws()
                except RequestError as e:
                    logger.info(e)
                    if e.reqId != -1:
                        self.__update_quotes_status(
                            error=True,
                            status_code=e.code,
                            status_text=e.message)
                        self.__pbar.update(incr=1)
                    continue
                else:
                    self.__write_quotes_to_db(bar_data)
                    self.__update_quotes_status(error=False)
                    self.__pbar.update(incr=1)
        except ExitSignalDetectedError:
            pass
        else:
            logger.info(
                f"Finished downloading historical data for universe "
                f"'{universe}'")
        finally:
            self.__tws_connector.disconnect()

    def __get_contracts(self, universe: str) -> List[int]:
        return self.__universes_db_connector.get_universe_members(
            universe=universe)

    def __get_contract_data(self) -> None:
        filters = {'contract_id': self.__contract_id}
        return_columns = ['broker_symbol', 'exchange', 'currency']
        contract_data = self.__contracts_db_connector.get_contracts(
            filters=filters,
            return_columns=return_columns)
        self.__contract_data = contract_data[0]

    def __get_contract_status(self) -> None:
        self.__quotes_status = self.__quotes_status_db_connector.get_quotes_status(
            contract_id=self.__contract_id)

    def __calculate_dates(self) -> None:
        """Calculate 'from' and 'to' dates for quotes to be downloaded"""

        if self.__quotes_status['status_code'] == 0:
            self.__calculate_dates_for_new_contract()
        elif self.__quotes_status['status_code'] == 1:
            self.__calculate_dates_for_existing_contract()
        else:
            self.__contract_has_error_status()
        logger.debug(f"Calculated 'duration' as: {self.__duration}, "
                     f"'from' as: {self.__quotes_from}, "
                     f"'till' as: {self.__quotes_till}")

    def __calculate_dates_for_new_contract(self) -> None:
        self.__duration = "15 Y"
        from_date = date.today() - timedelta(days=5479)  # 15 years
        self.__quotes_from = from_date.strftime('%Y-%m-%d')
        self.__quotes_till = date.today().strftime('%Y-%m-%d')

    def __calculate_dates_for_existing_contract(self) -> None:
        # Get config constants
        REDOWNLOAD_DAYS = int(self.__config_reader.get_config_value_single(
            section="quotes",
            option="redownload_days"))
        OVERLAP_DAYS = int(self.__config_reader.get_config_value_single(
            section="quotes",
            option="overlap_days"))

        start_date = (self.__quotes_status['daily_quotes_requested_till'])
        end_date = date.today().strftime('%Y-%m-%d')
        ndays = np.busday_count(start_date, end_date)
        if ndays <= REDOWNLOAD_DAYS:
            message = (
                f"Skipping {self.__contract_data['broker_symbol']}, existing "
                f"data is only {ndays} days old.")
            logger.debug(message)
            raise QuotesDurationError(message)
        if ndays > 360:
            message = (
                f"Skipping {self.__contract_data['broker_symbol']}, existing "
                f"data is already {ndays} days old.")
            logger.debug(message)
            raise QuotesDurationError(message)
        ndays += OVERLAP_DAYS
        self.__duration = str(ndays) + " D"
        self.__quotes_from = self.__quotes_status['daily_quotes_requested_from']
        self.__quotes_till = end_date

    def __contract_has_error_status(self) -> None:
        message = f"{self.__contract_data['broker_symbol']} already has error status. Skipped."
        logger.debug(message)
        raise ContractHasErrorStatusError(message)

    def __get_quotes_from_tws(self) -> BarDataList:
        quotes = self.__tws_connector.download_historical_quotes(
            symbol=self.__contract_data['broker_symbol'],
            exchange=self.__contract_data['exchange'],
            currency=self.__contract_data['currency'],
            duration=self.__duration)
        return quotes

    def __update_quotes_status(
            self,
            error: bool,
            status_code: int = 0,
            status_text: str = "") -> None:
        if error:
            self.__quotes_status_db_connector.update_quotes_status(
                contract_id=self.__contract_id,
                status_code=status_code,
                status_text=status_text,
                daily_quotes_requested_from="NULL",
                daily_quotes_requested_till="NULL")
        else:
            self.__quotes_status_db_connector.update_quotes_status(
                contract_id=self.__contract_id,
                status_code=1,
                status_text="Successful",
                daily_quotes_requested_from=self.__quotes_from,
                daily_quotes_requested_till=self.__quotes_till)

    def __write_quotes_to_db(self, bar_data: BarDataList) -> None:
        quotes = []
        for bar in bar_data:
            quote = (
                self.__contract_id,
                bar.date.strftime('%Y-%m-%d'),
                bar.open,
                bar.high,
                bar.low,
                bar.close,
                bar.volume)
            quotes.append(quote)
        self.__quotes_db_connector.insert_quotes(quotes=quotes)


class QuotesDurationError(Exception):
    """Problem occured with the 'from' and 'to' dates for downloading"""

    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)


class ContractHasErrorStatusError(Exception):
    """Contract has had an error on previous quotes download attempt"""
# TODO

    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)
